%******************************************************************
%
% Fast QZ for companion pencils
% (implementation of the algorithm in "Implicit QR for companion
% pencils", by P. Boito, Y. Eidelman and L. Gemignani).
%
% This code computes the roots of a polynomial P by forming the
% associated companion pencil and computing its eigenvalues.
% Quasiseparable structure is being used.
%
% This is written as a script, so the polynomial P 
% must be defined at the beginning, or be already in the workspace.
% Test examples are found in QZ_tests.m and in QZvsQR.m.
%
% The output includes vectors adiag and bdiag containing the
% diagonal entries of the two matrices A and B after the QZ 
% process has converged.
% The roots of P (eigenvalues of the companion pencil) can 
% be computed as adiag./bdiag.
%
% The variable total_it counts the total number of iterations,
% so that total_it/n gives the average number of iterations
% per eigenvalue.
%
%
% Last modified July 4, 2014.
%
% Please send your comments to paola.boito@unilim.fr.
%
%******************************************************************


% P should usually be normalized (if not already done):
 P=P/norm(P);

n=length(P)-1;
tol=1e-15;
total_it=0;
% find generators for matrix pencil
[sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=companion_generators(P);  % or
% [sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=companion_generators_col(P);

% recover A and B
A=recover_A(sigmaa,gv,hv,bv,z,w);
B=recover_B(db,gu,hu,bu,p,q);
tol_a=eps*(sqrt(n)+norm(z)*norm(w));
tol_b=eps*(sqrt(n)+norm(p)*norm(q));
% save A and B for later purposes
A0=A;
B0=B;
% initialize storage vectors for eigenvalues
adiag=zeros(n,1);
bdiag=zeros(n,1);  
% now let's use QZ to compute the eigenvalues of the pencil (A,B)

% perform deflation on A and on B:
% determine the indices kin and kfin 
% where the QZ process should start and then stop

kfin=n;
numit=0;

while kfin>1 


Bdefl=1;
while Bdefl==1
k=kfin-1;

while (k>0)&&(abs(db(k))>tol)
    k=k-1;
end
if k>0  % now k is the position of the zero diagonal entry
    fprintf('Performing deflation on B...')
    for j=k:kfin-1
    zz=[gu(j)*hu(j+1);db(j+1)];
    G=planerot(zz);
    % TO DO: apply structured QZ to push the zero entry to the bottom of B 
    
    B(j:j+1,:)=G*B(j:j+1,:);
    A(j:j+1,:)=G*A(j:j+1,:);
    xx=A(j+1,j-1:j);
    G=planerot(xx');
    B(:,j-1:j)=B(:,j-1:j)*fliplr(G');
    A(:,j-1:j)=A(:,j-1:j)*fliplr(G');
    end
    xx=A(kfin,kfin-1:kfin);
    G=planerot(xx');
    B(:,kfin-1:kfin)=B(:,kfin-1:kfin)*fliplr(G');
    A(:,kfin-1:kfin)=A(:,kfin-1:kfin)*fliplr(G');
    bdiag(kfin)=0;
    adiag(kfin)=A(kfin,kfin);
    kfin=kfin-1;
else
    Bdefl=0;
end
end
    
%
% deflation on A:     
% choose kfin    
j=kfin;
% several convergence criteria are possible here...
%while (j>1)&&(abs(A(j,j-1))<tol)
while (j>1)&&(abs(sigmaa(j-1))<eps*min(abs(gv(j,:)*hv(:,j)-z(j)*w(j)')+abs(gv(j-1,:)*hv(:,j-1)-z(j-1)*w(j-1)'),1))
%while (j>1)&&(abs(sigmaa(j-1))<eps*(abs(gv(j,:)*hv(:,j)-z(j)*w(j)')+abs(gv(j-1,:)*hv(:,j-1)-z(j-1)*w(j-1)')))
    fprintf('deflation\n')
    adiag(j)=gv(j,:)*hv(:,j)-z(j)*w(j)';
    bdiag(j)=db(j);
    numit=0;
    j=j-1;
end
kfin=j;
    
% choose kin with single subdiag criterion   
kin=kfin-1;
%while (kin>1)&&(abs(A(kin,kin-1))>tol)
while (kin>1)&&(abs(sigmaa(kin-1))>eps*min(abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')+abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)'),1))
%while (kin>1)&&(abs(sigmaa(kin-1))>eps*(abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')+abs(gv(kin,:)*hv(:,kin)-z(kin)*w(kin)')))    
    kin=kin-1;
end

if kin>0
% run QZ


% now do the magic trick on A (detection of starting point for iteration), increase kin if necessary
j=kfin-1;

if numit<10
xx=[db(kfin)*(gv(j,:)*hv(:,j)-z(j)*w(j)')-(gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)')*db(j);db(kfin)*sigmaa(j)];
else
    fprintf('exceptional shift\n')
    xx=rand(2,1)+1i*rand(2,1);
end
G=planerot(xx);    

if j>1 % do the magic trick, otherwise we already know kin=1
yy=G*[sigmaa(j-1);0];

while (j>kin)&&(abs(yy(2))>tol)   
    j=j-1; % candidate to become kin
% choose shift 
if numit<10
xx=[db(kfin)*(gv(j,:)*hv(:,j)-z(j)*w(j)')-(gv(kfin,:)*hv(:,kfin)-z(kfin)*w(kfin)')*db(j);db(kfin)*sigmaa(j)];
G=planerot(xx);
end
% check that column j-1 exists and compute yy
if j>1
    yy=G*[sigmaa(j-1);0];
end

end   
% reinitialize counter after exceptional shift
if numit>=10
    numit=0;
end
% update kin
kin=j;
else
    kin=1;
end

% perform structured QZ from kin to kfin

[sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q]=structured_QZ_infin(sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q,kin,kfin,G);

numit=numit+1;
total_it=total_it+1;

display('.')

end

end

adiag(1)=gv(1,:)*hv(:,1)-z(1)*w(1)';
bdiag(1)=db(1);